﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using isosoft.root;
using System.Globalization;

namespace TrieApp.model
{
	public class Trie : XObjectNotifier
	{
		static Trie		_instance		= null;

		CharDictionary	_dictionary	= new CharDictionary();
		TrieNodeList	_nodes		= new TrieNodeList();
		TrieNodeList	_root		= new TrieNodeList();


		public static Trie Instance
		{
			get
			{
				if(_instance == null)
					_instance = new Trie();

				return _instance;
			}
		}

		public static Trie GetInstance()
		{
			return _instance;
		}

		public Trie()
		{
			init_instance();
		}


		public CharDictionary Dictionary
		{
			get {  return _dictionary; }
		}

		public TrieNodeList RootNodes
		{
			get {  return _nodes; }
		}

		public TrieNodeList Root
		{
			get { return _root; }
		}


		void init_instance()
		{
			// insert alphabet
			InsertDictionaryNodes('a', 'z');

			// insert numbers and signes ('!' to '@' (this range includes 0-9 digits)
			InsertDictionaryNodes('\x21', '\x40');

			InitializeFirstLevel();
			TrieNode	root	= new TrieNode(new CharDictionaryItem('■', null, null), null, null);

			root.IsExpanded	= true;
			_root.Add(root);

			root.Children	= _nodes;
		}


		void InsertDictionaryNodes(char min, char max)
		{
			CharDictionaryItem	prevNode	= null,
								newNode		= null;

			// insert dictionary items
			for(char c = min; c <= max; c++)
			{
				newNode = new CharDictionaryItem(c, prevNode, null);
				_dictionary.Add(newNode);
				
				if(prevNode != null)
					prevNode.Next	= newNode;

				prevNode	= newNode;
			}
		}

		void InitializeFirstLevel()
		{
			_nodes.Clear();
			foreach(CharDictionaryItem item in _dictionary)
			{
				_nodes.AddTail(item, false);
			}
		}

		CharDictionaryItem GetDictionaryItem(char c, bool notifyChanges)
		{
			CharDictionaryItem	dicoItem	= _dictionary[c];

			if (dicoItem == null)
			{
				dicoItem = _dictionary.AddTail(c, notifyChanges);
				if(notifyChanges)
					NotifyPropertyChanged(() => Dictionary);
			}

			return dicoItem;
		}

		public void AddString(string str, object finalValue, bool notifyChanges)
		{
			if( string.IsNullOrEmpty(str))
				return;

			// level 0
			char				c			= str[0];
			CharDictionaryItem	dicoItem;
			TrieNode			firstNode	= _nodes[c],
								node;
			int					ndx,
								len			= str.Length;

			if(firstNode == null)
			{
				dicoItem	= GetDictionaryItem(c, notifyChanges);
				firstNode	= _nodes.AddTail(dicoItem, notifyChanges);
			}

			for(ndx = 1; ndx < len; ndx++)
			{
				c			= str[ndx];
				node		= firstNode.Children[c];
				dicoItem	= GetDictionaryItem(c, notifyChanges);

				if(node == null)
				{
					node = firstNode.Children.AddTail(dicoItem, notifyChanges);

					if(notifyChanges)
					{
						NotifyPropertyChanged(() => RootNodes);
						NotifyPropertyChanged(() => Root);
					}
				}
					

				if(ndx >= len -1)
				{
					node.IsEndOfWord	= true;
					node.Value			= finalValue;
				}
				firstNode = node;
			}

			if(notifyChanges)
				NotifyPropertyChanged(() => AllStrings);
			// NotifyPropertyChanged(() => Dictionary);
		}

		public object SearhString(string str)
		{
			if(string.IsNullOrEmpty(str))
				return null;

			int			ndx			= 0,
						len			= str.Length;
			char		c			= str[0];
			TrieNode	firstNode	= _nodes[c],
						node		= null;

			if(firstNode == null)
				return null;

			for(ndx = 1; ndx < len; ndx++)
			{
				c		= str[ndx];
				node	= firstNode.Children[c];

				if( node == null)
					break;

				firstNode	= node;
			}

			if(node == null || ndx < len || ! node.IsEndOfWord)
				return null;

			return node.Value;
		}

		/// <summary>
		/// find words starting with the given string
		/// </summary>
		/// <param name="str">the string</param>
		/// <returns>the list of strings starting with str</returns>
		public List<string> StartsWith(string str)
		{
			List<String>	list	= new List<string>();

			if(string.IsNullOrEmpty(str))
				return list;

			char		c		= str[0];
			TrieNode	node0	= this._nodes[c];

			if(node0 == null)
				return list;

			List<string>	strings	= node0.Strings;

			foreach(string s in strings)
			{
				if(s.StartsWith(str, true, CultureInfo.CurrentCulture))
					list.Add(s);
			}

			return list;
		}

		/// <summary>
		/// find strings containing the given string
		/// </summary>
		/// <param name="str">the string to find</param>
		/// <returns>strings containing str</returns>
		public List<string> Containing(string str)
		{
			List<String>	list	= new List<string>();

			if(string.IsNullOrEmpty(str))
				return list;

			foreach(TrieNode node in this._nodes)
			{
				var	strings = from sx in node.Strings where( sx.Contains(str)) select sx;

				foreach(string s in strings)
				{
					list.Add(s);
				}
			}

			return list;
		}

		/// <summary>
		/// get the list of all trie's strings
		/// </summary>
		public List<string> AllStrings
		{
			get
			{
				List<String> list = new List<string>();

				foreach (TrieNode node in this._nodes)
				{
					var strings = from sx in node.Strings select sx;

					foreach (string s in strings)
					{
						list.Add(s);
					}
				}

				return list;
			}
		}


		void AddChildNodes(List<List<TrieNode>> targetList, TrieNode node)
		{
			if(targetList == null || node == null)
				return;

			List<TrieNode>	neighbourList	= new List<TrieNode>();
			TrieNode		neighbour		= node.Next;

			neighbourList.Add(node);

			while(neighbour != null)
			{
				neighbourList.Add(neighbour);
				neighbour	= neighbour.Next;
			}

			if(neighbourList.Count >0)
				targetList.Add(neighbourList);

			foreach(TrieNode node2 in neighbourList)
			{
				if (node2.Children != null && node2.Children.Count > 0)
					AddChildNodes(targetList, node2.Children[0]);
			}
		}


		public void Sort()
		{
			_nodes.SortNodes();
		}

		public List<List<TrieNode>> NodesByDepthLevel
		{
			get
			{
				List<List<TrieNode>>	list	= new List<List<TrieNode>>();
				if(_nodes == null || _nodes.Count <= 0)
					return list;

				AddChildNodes(list, _nodes[0]);
				return list;
			}
		}


		internal void ClearAll(bool notifyChanges)
		{
			_nodes.Clear();
			_dictionary.Clear();

			if(notifyChanges)
			{
				NotifyPropertyChanged(() => RootNodes);
				NotifyPropertyChanged(() => Root);
			}
		}

		Task<int> ParseStringTask(string str, int minWordLength, bool notifyChanges)
		{
			Task<int> task = Task.Factory.StartNew<int>(
				() =>
					{
						this.ClearAll(notifyChanges);

						// split text word
						string[] words = str.Split(new string[]
								{ 
									/*"\0",*/	" ",	"\r\n",	"(",	")",	",",	".",	"'",	"\"", 
									":",	"/",	"‘",	"’",	"«",	"»",	"-",	"+",	"=",
									"*",	"[",	"]",	"{",	"}",	";",	"&",	"<",
									">",	"|",	"\\",	"`",	"^",	"…",	"\t"	
								}
										, StringSplitOptions.RemoveEmptyEntries);
						string	word;

						foreach (string w in words)
						{
							word = w.Trim();

							if (word.Length >= minWordLength)
							{
								this.AddString(word, null, notifyChanges);
							}
						}

						return this._nodes.Count;
					});

			return task;
		}


		public async Task<int> LoadFromStringAsync(string str, int minWordLength, bool notifyChanges)
		{
			int		nodesCount	= await ParseStringTask(str, minWordLength, notifyChanges);

			return nodesCount;
		}


	}
}
